/* file csputil.c
 *
 * NIST PIV Demo CSP -- Cryptographic Service Provider
 *
 * Designed and Developed by Electrosoft (www.electrosoft-inc.com) based on requirements
 * from the National Institute of Standards and Technology (NIST).
 *
 * This file codes and documents many utility functions.
 *
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#undef UNICODE                  // ## Not Yet
#include <windows.h>
#include <wincrypt.h>
#include <cspdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <winscard.h>
#include "csp-debug.h"
#include "cspdefs.h"
#include "zlib.h"
#include "BERtlv.h"

#ifdef ISO24727
#include "iso24727api.h"
#include "EncodingRoutines.h"

char* DIDNAME_APPLICATION_PIN			= "Application PIN";

#else									// PIV
#include "piv.h"
#endif

#include "csp.h"

#define MAX_BUFFER_SIZE 65536
#define MAX_READERS 20
#define MAX_UNCOMPRESS_BUFFER_SIZE		8*1024

#define GZIP_COMPRESSION				0x01

//extern BOOL cardConnected;
BOOL cardConnected = FALSE;

ALGORITHM Algs[] = {
	// MINIMAL INFORMATION.....
    {CALG_DES,DES_BITS,DES_NAME,DES_MIN_BITS, DES_MAX_BITS,NULL,0},
    {CALG_3DES,DES3_BITS,DES3_NAME,DES3_MIN_BITS, DES3_MAX_BITS,NULL,0},
    {CALG_RC2,RC2_BITS,RC2_NAME,RC2_MIN_BITS, RC2_MAX_BITS,NULL,0},
    {CALG_RC4,RC4_BITS,RC4_NAME, RC4_MIN_BITS, RC4_MAX_BITS,NULL,0}
}; /**< Supported algorithms database.*/

int piv_login2(PROV_CTX *pProvCtx, char *pindata, int pinlen, SCARDHANDLE cardHandle)
{
	char debug[50];
	int ii;
	PIV_RV rv;

#ifdef ISO24727
	BYTE CardholderPIN[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
	DIDAuthenticationData didAuthData = {sizeof(CardholderPIN), CardholderPIN};
	DIDScope didLocalScope;
	DIDName didName;

	didLocalScope.choiceId = DIDSCOPE_LOCAL;
	didLocalScope.a.local = 0;

	didName.choiceId = DIDNAME_PLAINNAME;
	didName.a.plainName = StringOcts(DIDNAME_APPLICATION_PIN);

	for (ii=0; ii<pinlen; ii++)
		didAuthData.octs[ii] = pindata[ii];

	rv = CardApplicationStartSession(cardHandle, didLocalScope, didName, &didAuthData, NULL);
	FreeDIDName(&didName);
#else
	PIV_Byte authData[15];
	PIV_Byte authDataLen;

	authData[0] = 0x67;
	authData[1] = 13;

	authData[2] = 0x81;
	authData[3] = 8;
	authData[4] = 0xff;
	authData[5] = 0xff;
	authData[6] = 0xff;
	authData[7] = 0xff;
	authData[8] = 0xff;
	authData[9] = 0xff;
	authData[10]= 0xff;
	authData[11]= 0xff;

	authData[12]= 0x83;
	authData[13]= 1;
	authData[14] = 0x80;

	for (ii=0; ii<pinlen; ii++)
		authData[4+ii] = pindata[ii];

	authDataLen = 15;

	sprintf(debug, "Auth data len -> %d\n", authDataLen);
	DEBUG(3, debug);

	rv = pivLogIntoCardApplication((PIV_CARDHANDLE) cardHandle, authData, authDataLen);
#endif

	if (rv != CSP_S_OK)
	{
		sprintf(debug, "PIN Verification failed!!!!#####@@@@@@. %d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}
	DEBUG(3, "\nLogged into card application\n");
	return TRUE;
}

BOOL piv_login(PROV_CTX *pProvCtx, char *pp_pin, DWORD keySpec, SCARDHANDLE cardHandle)
{
    BOOL rv;
    char pin[30]; /* The entered pin.*/

    /**  - We assume the pin is incorrect if no pin is entered yet.*/
    rv = FALSE;

    /** - Before everything, if pp_pin is NULL, and caching is permitted,*/
    if((pp_pin == NULL) && (!pProvCtx->nocache))
    {
        /**   - Try cached PIN if exists.*/
        switch(keySpec)
        {
            case AT_SIGNATURE:
                if(pProvCtx->cachedSigPin != NULL)
                {
                    DEBUG(3, "Cached signature PIN used.\n");
                    /** - Log into the card once. */
					rv = piv_login2(pProvCtx, pProvCtx->cachedSigPin, strlen(pProvCtx->cachedSigPin), cardHandle);
					if (rv == TRUE)
                    /** - If logged, ok.*/
                    {
                        return TRUE;
                    }
                    else
                    /** - Else free cache.*/
                    {
                        HeapFree(pProvCtx->heap, 0, pProvCtx->cachedSigPin);
                        pProvCtx->cachedSigPin = NULL;
                        DEBUG(3, "Wrong cached signature PIN.\n");
						return FALSE;
                    }
                }
                break;
            case AT_KEYEXCHANGE:
                if(pProvCtx->cachedKeyExchangePin != NULL)
                {
                    DEBUG(4, "Cached key exchange PIN used.\n");
                    /** - Log into the card once. */
					rv = piv_login2(pProvCtx, pProvCtx->cachedSigPin, strlen(pProvCtx->cachedSigPin), cardHandle);
					if (rv == TRUE)
                    /** - If logged, ok.*/
                    {
                        return TRUE;
                    }
                    else
                    /** - Else free cache.*/
                    {
                        HeapFree(pProvCtx->heap, 0, pProvCtx->cachedKeyExchangePin);
                        pProvCtx->cachedKeyExchangePin = NULL;
                        DEBUG(3, "Wrong cached key exchange PIN.\n");
						return FALSE;
					}
                }
				break;
		}
	}


	if(pp_pin != NULL)
    {
        //Copy the eight first characters in the pin to be passed.
        strncpy(pin,pp_pin,8);
		pin[8] = '\0';					// add null-terminator -- if PIN is
										// less than 8 chars, null-terminator
										// is automatically added
    }

    /** - Log into the card once. */
	rv = piv_login2(pProvCtx, pin, strlen(pin), cardHandle);
	DEBUG(3, "PIV_login2 came back\n");

	return rv;
}

BOOL simpleSignData(PROV_CTX *pProvCtx, DWORD dwKeySpec, BYTE *pData, DWORD dataLen,
              BYTE *pSignature, DWORD *pSignatureLen, SCARDHANDLE cardHandle, char* cachedPIN)
{
	char debug[500];
	unsigned int ii;
	PIV_RV rv;

#ifdef ISO24727
	DIDScope didLocalScope;
	DIDName didName;
	MessageBuffer messageBuffer;
	SignatureBuffer signatureBuffer;
	char szDIDName[64];
	unsigned char message[128];			// hardcode for RSA-1024 key length for now

	if(dwKeySpec == AT_SIGNATURE)
		strcpy(szDIDName, "PIV Card Application Digital Signature Key");
	else
		strcpy(szDIDName, "PIV Authentication Key");

	sprintf(debug, "Key ref name is '%s'\n", szDIDName);
	DEBUG(3, debug);

	for (ii=0; ii < dataLen; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", pData[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

	if( !piv_login(pProvCtx, cachedPIN, AT_KEYEXCHANGE, cardHandle) )
	{
		DEBUG(3, "Login just before calling crypt operation failed\n");
        SetLastError(SCARD_W_WRONG_CHV);
        return FALSE;
	}

	didLocalScope.choiceId = DIDSCOPE_LOCAL;
	didLocalScope.a.local = 0;

	didName.choiceId = DIDNAME_PLAINNAME;
	didName.a.plainName = StringOcts(szDIDName);

	message[0] = 0x00;
	message[1] = 0x01;
	memset(&message[2], 0xff, 91);
	message[93] = 0x00;
	memcpy(&message[94], pData, dataLen);

	messageBuffer.octetLen = sizeof(message);
	messageBuffer.octs = message;

	// compute signature
	rv = Sign(cardHandle, didLocalScope, didName, messageBuffer, &signatureBuffer);
	FreeDIDName(&didName);

	if (rv == CSP_S_OK) {

		// make sure there's enough space in the output buffer to copy the 
		// signature
		if (signatureBuffer.octetLen <= *pSignatureLen) {

			memcpy(pSignature, signatureBuffer.octs, signatureBuffer.octetLen);
			*pSignatureLen = signatureBuffer.octetLen;
		}
		else
			rv = CSP_S_INSUFFICIENT_BUFFER;

		FreeSignatureBuffer(&signatureBuffer);
	}
#else
	PIV_Byte algID = 0x06; // Hardcoded for now
	PIV_Byte keyReference;

	unsigned char alginput[200];

	alginput[0] = 0x7c;
	alginput[1] = 0x81;
	alginput[2] = 0x85;
	alginput[3] = 0x81;
	alginput[4] = 0x81;
	alginput[5] = 0x80;
	alginput[6] = 0x00;
	alginput[7] = 0x01;
	memset(&alginput[8], 0xff, 91);
	alginput[99] = 0x00;

	memcpy(&alginput[100], pData, dataLen);

	alginput[134] = 0x82;
	alginput[135] = 0x00;

	if(dwKeySpec == AT_SIGNATURE)
	{
		keyReference = 0x9C;
	}
	else
	{
		keyReference = 0x9A;
	}
	sprintf(debug, "Key ref is %02X ", keyReference);
	DEBUG2(3, debug, 14);
	for (ii=0; ii < dataLen; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", alginput[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

	if( !piv_login(pProvCtx, cachedPIN, AT_KEYEXCHANGE, cardHandle) )
	{
		DEBUG(3, "Login just before calling crypt operation failed\n");
        SetLastError(SCARD_W_WRONG_CHV);
        return FALSE;
	}

    rv = pivCrypt((PIV_CARDHANDLE) cardHandle, algID, keyReference,
		alginput, dataLen+102 , pSignature, pSignatureLen);
#endif

	if (rv != CSP_S_OK)
	{
		*pSignatureLen = 0;
		sprintf(debug, "Crypt Operation failed!!!!#####@@@@@@. %d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}

	return TRUE;
}

BOOL decryptData(PROV_CTX *pProvCtx, BYTE *pData, DWORD dataLen,
              BYTE *pSignature, DWORD *pSignatureLen, SCARDHANDLE cardHandle, char* cachedPIN)
{
	char debug[500];
	unsigned int ii;
	PIV_RV rv;

#ifdef ISO24727
	DIDScope didLocalScope;
	DIDName didName;
	CipherBuffer cipherText, plainText;
	char szDIDName[] = "PIV Authentication Key";

	sprintf(debug, "Key ref name is '%s'\n", szDIDName);
	DEBUG(3, debug);

	if (dataLen < 128)					// hardcode for RSA 1024 for now
	{
		DEBUG(3, "Not enough data to decrypt!\n");
        return FALSE;
	}

	for (ii=0; ii < 128; ii++)			// only care about first 128 bytes -- 
	{									// remaining is MS Blob header etc....
		sprintf(&debug[ii*2], "%02X ", pData[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

	if( !piv_login(pProvCtx, cachedPIN, AT_KEYEXCHANGE, cardHandle) )
	{
		DEBUG(3, "Login just before calling crypt operation failed\n");
        SetLastError(SCARD_W_WRONG_CHV);
        return FALSE;
	}

	didLocalScope.choiceId = DIDSCOPE_LOCAL;
	didLocalScope.a.local = 0;

	didName.choiceId = DIDNAME_PLAINNAME;
	didName.a.plainName = StringOcts(szDIDName);

	cipherText.octetLen = 128;			// hardcode for RSA 1024 for now
	cipherText.octs = pData;

	// decrypt data
	rv = Decipher(cardHandle, didLocalScope, didName, cipherText, &plainText);
	FreeDIDName(&didName);

	if (rv == CSP_S_OK) {

		// make sure there's enough space in the output buffer to copy the 
		// decrypted data
		if (plainText.octetLen <= *pSignatureLen) {

			memcpy(pSignature, plainText.octs, plainText.octetLen);
			*pSignatureLen = plainText.octetLen;
		}
		else
			rv = CSP_S_INSUFFICIENT_BUFFER;

		FreeCipherBuffer(&plainText);
	}
#else
	PIV_Byte algID = 0x06; // Hardcoded for now
	PIV_Byte keyReference = 0x9a;

	unsigned char alginput[200];

	alginput[0] = 0x7c;
	alginput[1] = 0x81;
	alginput[2] = 0x85;
	alginput[3] = 0x81;
	alginput[4] = 0x81;
	alginput[5] = 0x80;
	memcpy(&alginput[6], pData, dataLen-10); // 10 for the MS Blob header etc....
	alginput[134] = 0x82;
	alginput[135] = 0x00;

	sprintf(debug, "Key ref is %02X ", keyReference);
	DEBUG2(3, debug, 14);
	for (ii=0; ii < (dataLen-10); ii++)
	{
		sprintf(&debug[ii*2], "%02X ", alginput[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

	if( !piv_login(pProvCtx, cachedPIN, AT_KEYEXCHANGE, cardHandle) )
	{
		DEBUG(3, "Login just before calling crypt operation failed\n");
        SetLastError(SCARD_W_WRONG_CHV);
        return FALSE;
	}

    rv = pivCrypt((PIV_CARDHANDLE) cardHandle, algID, keyReference,
		alginput, 136 , pSignature, pSignatureLen);
#endif

	if (rv != CSP_S_OK)
	{
		*pSignatureLen = 0;
		sprintf(debug, "Crypt Operation failed!!!!#####@@@@@@. %d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}

	return TRUE;
}

BOOL getAlgorithm(ALG_ID algId, ALGORITHM *algorithm)
{
	char debug[30];
    int tableLength = 0; /* The translation table lenth, computed here.*/
    int i = 0; /* Iterator.*/

    tableLength = sizeof(Algs) / sizeof(ALGORITHM);

	sprintf(debug, "Algorithm table length is ->%d\n", tableLength);
	DEBUG(3, debug);
	sprintf(debug, "Alg Id coming in is ->%d\n", algId);
	DEBUG(3, debug);

	*algorithm = Algs[algId];
	DEBUG(3, "Sending algo struct\n");
	return TRUE;

}


int piv_connect(PROV_CTX *pProvCtx, SCARDHANDLE *cardHandle) {

	unsigned char debug[100];

	sprintf(debug, "Cardconnected is %d\n", cardConnected);
	DEBUG(3, debug);
	sprintf(debug, "Cardconnected var pointer is %x\n", &cardConnected);
	DEBUG(3, debug);

	if( (*cardHandle == 0) && (!cardConnected))
	{
		PIV_CARDHANDLE hCard=0;
		PIV_RV rv;

#ifdef ISO24727
		CardApplicationPathA cardAppPathRequest;
		CardApplicationPathSet* cardAppPathResultSet;
		BYTE PIVAID[] = {0xA0, 0x00, 0x00, 0x03, 0x08, 0x00, 0x00, 0x10, 0x00, 0x01, 0x00};

		// set up card app path request to look for PIV card app in local
		// readers
		cardAppPathRequest.pathSecurity = NULL;
		cardAppPathRequest.channelHandle = NULL;
		InitOctsNoAlloc(&cardAppPathRequest.contextHandle);
		InitOctsNoAlloc(&cardAppPathRequest.iFDName);
		cardAppPathRequest.slotIndex = NULL;
		ByteOctsNoAlloc(&cardAppPathRequest.cardApplication, PIVAID, sizeof(PIVAID));

		// find path to PIV card application
		rv = CardApplicationPath(cardAppPathRequest, &cardAppPathResultSet);
		FreeCardApplicationPathA(&cardAppPathRequest);
		
		if (rv == CSP_S_OK) {
			CardApplicationPathA *pcardAppPath;

			pcardAppPath = (CardApplicationPathA*) FIRST_LIST_ELMT(cardAppPathResultSet);

			// connect to PIV card application
			rv = CardApplicationConnect(*pcardAppPath, FALSE, &hCard);

			FreeCardApplicationPathSet(cardAppPathResultSet);
		}
#else
		LPSTR readers, chosenReader;
		DWORD readersLen;
		int nReaders=0, readerNumber;
		int iList[MAX_READERS]; /* all readers after MAX_READERS will be ignored*/
		PIV_ULong32 bufferLength, i;
		PIV_Byte connectionDescription[500] = {0x7f, 0x21, 0x04, 0x81, 0x00, 0x90, 0x00};
		PIV_ULong32 cdLength=500;

		bufferLength = 0;

		readers	= NULL;

		rv = pivConnect(0, connectionDescription, &cdLength, &hCard);
		if (rv != CSP_S_OK) {
			DEBUG (3, "PC/SC appears to be working, but a list of readers could not be retrieved.\n");
			return rv;
		}

		readers = malloc(sizeof(char)*cdLength);
		readersLen = cdLength;
		memcpy (readers, connectionDescription, cdLength);

		for (i = 0; i < cdLength - 1; i++) {
			++nReaders;
			iList[nReaders] = i;
			while (readers[++i] != 0);
		}

		DEBUG(3, "Came out from going thru the readers after calling PIVConnect\n");
		readerNumber = 1;

		chosenReader = &(readers[iList[readerNumber]]);
		DEBUG(3, "Attempting to connect...\n");

		connectionDescription[0] = 0x7f;
		connectionDescription[1] = 0x21;
		connectionDescription[2] = (PIV_Byte)strlen(chosenReader)+4;
		connectionDescription[3] = 0x81;
		connectionDescription[4] = (PIV_Byte)strlen(chosenReader);
		memcpy(connectionDescription+5, chosenReader, connectionDescription[4]);
		connectionDescription[5+connectionDescription[4]] = 0x90;
		connectionDescription[5+connectionDescription[4]+1] = 0;

		cdLength = connectionDescription[2] + 2;

		free(readers);

		rv = pivConnect(0, connectionDescription, &cdLength, &hCard);
#endif

		if (rv != CSP_S_OK)
		{
			if (rv == CSP_S_CONNECTION_LOCKED)
			{
				DEBUG(3, "\nCard is locked by another application.\n");
			}
			else
			{
				sprintf(debug, "\nConnection to card failed. %d\n", rv);
				DEBUG(3, debug);
			}
			return FALSE;
		}
		*cardHandle = hCard;
		cardConnected = TRUE;

		DEBUG(3, "done.\nExclusive connection formed.\n");
	}

	return TRUE;
}

int piv_select(PROV_CTX *pProvCtx, SCARDHANDLE cardHandle)
{
	// When compiling for ISO/IEC 24727, we don't need to select the card app 
	// since this is done by CardApplicationConnect. When compiling for PIV, we
	// need to select the card app.
#ifndef ISO24727
	char debug[100];
	PIV_Byte pivAID[11] = {0xA0, 0, 0, 3, 8, 0, 0, 0x10, 0, 1, 0};
	PIV_Byte *propertyTemplate;
	PIV_ULong32 templateLength=0;
	PIV_RV rv;

	DEBUG (3, "Selecting the PIV Applet...\n");

	sprintf(debug, "Card handle for pivSelectCardApplication ->%02x\n", cardHandle);
	DEBUG(3, debug);
	rv = pivSelectCardApplication((PIV_CARDHANDLE) cardHandle, pivAID, 11, NULL, &templateLength);
	if (rv != CSP_S_INSUFFICIENT_BUFFER) {
		DEBUG (3, "could not select applet.\n");
		return rv;
	}
	propertyTemplate = malloc (sizeof(PIV_Byte)*templateLength);
	sprintf(debug, "Total bytes allocated ->%x\n", sizeof(PIV_Byte)*templateLength);
	DEBUG(3, debug);
	rv = pivSelectCardApplication((PIV_CARDHANDLE) cardHandle, pivAID, 11, propertyTemplate, &templateLength);
	free(propertyTemplate);
	DEBUG (3, "Applet successfully selected\n");
#endif

	return TRUE;
}

int GZipInflate(unsigned char *pSrc, unsigned long ulSrcSize, unsigned char **ppDest, unsigned long *pulDestSize)
{
	z_stream strm;
	unsigned char *pRealloc;
	unsigned char binBuffer[MAX_UNCOMPRESS_BUFFER_SIZE];
	unsigned long ulBytesUncompressed;
	int ret;

	// initialize output
	*ppDest = NULL;
	*pulDestSize = 0;

	if (ulSrcSize == 0)
		return Z_DATA_ERROR;

	// allocate inflate state
	strm.zalloc = Z_NULL;
	strm.zfree = Z_NULL;
	strm.opaque = Z_NULL;
	strm.avail_in = 0;
	strm.next_in = Z_NULL;

	ret = inflateInit2(&strm,47);		// 47 = gzip
	if (ret != Z_OK)
		return ret;

	strm.avail_in = ulSrcSize;
	strm.next_in = pSrc;

	// run inflate() on input until output buffer not full
	do {
		strm.avail_out = sizeof(binBuffer);
		strm.next_out = binBuffer;

		ret = inflate(&strm, Z_NO_FLUSH);
		//ASSERT(ret != Z_STREAM_ERROR);  // state not clobbered

		switch (ret) {
		case Z_NEED_DICT:
			ret = Z_DATA_ERROR;     // and fall through
		case Z_DATA_ERROR:
		case Z_MEM_ERROR:
			(void)inflateEnd(&strm);

			if (*ppDest)
			{
				free(*ppDest);
				*ppDest = NULL;
				*pulDestSize = 0;
			}

			return ret;
		}

		// copy uncompressed data to output buffer, increasing the buffer size
		// if necessary
		ulBytesUncompressed = sizeof(binBuffer) - strm.avail_out;
		if (ulBytesUncompressed > 0)
		{
			*pulDestSize += ulBytesUncompressed;
			pRealloc = (unsigned char*) realloc(*ppDest, *pulDestSize);

			if (pRealloc)
			{
				*ppDest = pRealloc;
				memcpy(*ppDest + *pulDestSize - ulBytesUncompressed, binBuffer, ulBytesUncompressed);
			}
			else
			{
				ret = Z_MEM_ERROR;
				break;
			}
		}
	} while (strm.avail_out == 0);

	// clean up and return
	(void)inflateEnd(&strm);
	ret = (ret == Z_STREAM_END) ? Z_OK : Z_DATA_ERROR;

	if (ret != Z_OK && *ppDest)
	{
		free(*ppDest);
		*ppDest = NULL;
		*pulDestSize = 0;
	}

	return ret;
}

BOOL loadUserKey(PROV_CTX *pProvCtx, KEY_INFO *pKeyInfo, SCARDHANDLE cardHandle)
{
	char debug[100];
	PIV_Byte *data, *pTlv, *pCert = NULL;
	PIV_ULong32 dataLength, ulTag, ulLen, ulCertSize, ulCertInfoMask;
	PIV_Byte ucTagLen, ucLenLen;
	PIV_RV rv;
	BOOL bCompressed = FALSE;

#ifdef ISO24727
	DataSetName dataSetName;
	DSIName dsiName;
	DSIContent dsiContent;
	char szDsiName[64];

	// For now, use PIV Auth key for both. Should Digital Signature key be used
	// for AT_SIGNATURE?
	if(pKeyInfo->dwKeySpec == AT_SIGNATURE )
    {
		DEBUG(3, "Going for SIGNATURE key\n");
		strcpy(szDsiName, "X.509 Certificate for PIV Authentication");
	}
	else // AT_KEYEXCHANGE
	{
		DEBUG(3, "Going for KEYEXCHANGE key\n");
		strcpy(szDsiName, "X.509 Certificate for PIV Authentication");
	}

	StringOctsNoAlloc(&dataSetName, szDsiName);

	// select the data set corresponding to the requested data object
	rv = DataSetSelect(cardHandle, dataSetName);
	FreeDataSetName(&dataSetName);

	if (rv != CSP_S_OK) {
		sprintf(debug, "Failed to select data set '%s' -> %d\n", szDsiName, rv);
		DEBUG(3, debug);
		return FALSE;
	}

	StringOctsNoAlloc(&dsiName, szDsiName);

	// read the data object
	rv = DSIRead(cardHandle, dsiName, &dsiContent);
	FreeDSIName(&dsiName);

	if (rv != CSP_S_OK) {
		sprintf(debug, "Error encountered during call to DSIRead ->%d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}

	dataLength = dsiContent.octetLen;
	data = malloc(dataLength);
	memcpy(data, dsiContent.octs, dataLength);

	FreeDSIContent(&dsiContent);
#else
	PIV_Byte oid[100];
	PIV_ULong32 oidLength;

	if(pKeyInfo->dwKeySpec == AT_SIGNATURE )
    {
		DEBUG(3, "Going for SIGNATURE key\n");
		memcpy (oid, "2.16.840.1.101.3.7.2.1.1", 24);
		oid[24]=0;
		oidLength = 24;
	}
	else // AT_KEYEXCHANGE
	{
		//2.16.840.1.101.3.7.2.5.0 - Card Authentication
		//2.16.840.1.101.3.7.2.1.0 - PIV Authentication
		DEBUG(3, "Going for KEYEXCHANGE key\n");

		//Using Card Authentication Certificate
		//memcpy (oid, "2.16.840.1.101.3.7.2.5.0", 24);

		//Using PIV Authentication Certificate
		memcpy (oid, "2.16.840.1.101.3.7.2.1.1", 24);

		oid[24]=0;
		oidLength = 24;
	}
	sprintf(debug, "Card handle for getdata ->%x\n", cardHandle);
	DEBUG(3, debug);

	rv = pivGetData((PIV_CARDHANDLE) cardHandle, oid, oidLength, NULL, &dataLength);
	if (rv != CSP_S_INSUFFICIENT_BUFFER) {
		sprintf(debug, "Error encountered during pivGetData 1st call ->%d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}

	data = malloc (sizeof(PIV_Byte)*dataLength);

	// read the data object
	rv = pivGetData((PIV_CARDHANDLE) cardHandle, oid, 3, data, &dataLength);
	if (rv != CSP_S_OK) {
		free(data);
		sprintf(debug, "Error encountered during pivGetData 2nd call ->%d\n", rv);
		DEBUG(3, debug);
		return FALSE;
	}
#endif

	pTlv = data;

	// parse TLV tags to get cert
	do
	{
		ulTag = getTagLength(pTlv, &ucTagLen, &ulLen, &ucLenLen);

		switch (ulTag)
		{
			case 0x70:
				pCert = pTlv + ucTagLen + ucLenLen;
				ulCertSize = ulLen;
				break;
			case 0x71:
				ulCertInfoMask = tlvIntegerValue(pTlv + ucTagLen + ucLenLen, ulLen);
				if (ulCertInfoMask & GZIP_COMPRESSION)
					bCompressed = TRUE;
				break;
		}

		pTlv += ucTagLen + ucLenLen + ulLen;
	} while (pTlv < data + dataLength);

	if (pCert == NULL)
	{
		free(data);
		sprintf(debug, "Unable to locate certificate in TLV tags\n");
		DEBUG(3, debug);
		return FALSE;
	}

	// uncompress cert
	if (bCompressed)
	{
		if (GZipInflate(pCert, ulCertSize, &pKeyInfo->data, &pKeyInfo->dataLen) != Z_OK)
		{
			free(data);
			sprintf(debug, "Unable to decompress certificate\n");
			DEBUG(3, debug);
			return FALSE;
		}
	}
	else
	{
		pKeyInfo->data = (PIV_Byte*) malloc(ulCertSize);
		memcpy(pKeyInfo->data, pCert, ulCertSize);
		pKeyInfo->dataLen = ulCertSize;
	}

	free(data);

	sprintf(debug, "Load User Key returning fine with datalength = %lu\n", pKeyInfo->dataLen);
	DEBUG(3, debug);

	return TRUE;
}

BOOL openPIVContainer(PROV_CTX *pProvCtx, SCARDHANDLE *cardHandle)
{
	char debug[100];
	DEBUG(3, "openPIVContainer executing\n");
	//pivConnect, pivSelectCardApplication
	if( !piv_connect(pProvCtx, cardHandle) )
	{
		return FALSE;
	}
	sprintf(debug, "PIV Connect Succedded handle = %02x\n", *cardHandle);
	DEBUG(3, debug);
	if( !piv_select(pProvCtx, *cardHandle) )
	{
		return FALSE;
	}
	DEBUG(3, "PIV Select Succedded\n");
    return TRUE;
}

BOOL closePIVContainer(PROV_CTX *pProvCtx, SCARDHANDLE *cardHandle)
{
	//pivDisconnect, pivLogoutOfCardApplication,
	PIV_RV rv;
	char debug[100];
	sprintf(debug, "Logout/Disconnect may happen, card handle-> %02x\n", *cardHandle);
	DEBUG(3, debug);
	if( *cardHandle != 0 )
	{
		/* DONT WANT TO LOGOUT FOR NOW - TODO
		rv = pivLogoutOfCardApplication(*cardHandle);
		if( rv != CSP_S_OK )
		{
			sprintf(debug, "pivLogoutOfCardApplication failed -> %d\n", rv);
			DEBUG(3, debug);
			return FALSE;
		}
		DEBUG(3, "PIV Logout Succedded\n");
		*/
#ifdef ISO24727
		rv = CardApplicationDisconnect(*cardHandle, NULL);
#else
		rv = pivDisconnect(*cardHandle);
#endif
		if( rv != CSP_S_OK )
		{
			sprintf(debug, "pivDisconnect failed -> %d\n", rv);
			DEBUG(3, debug);
			return FALSE;
		}
		DEBUG(3, "PIV Disconnect Succedded\n");
		*cardHandle = 0;
		cardConnected = FALSE;
	}
	return TRUE;
}

BOOL reverseBytesString(BYTE *pBytes, DWORD stringLen)
{
    BYTE byteBuffer;      /* A byte used to buffer during byte reversing.*/
    DWORD middle;       /* The length in byte of a middle of signature.*/
    unsigned int i;              /* iterator.*/

    middle = stringLen/2;
    for(i=0;i<middle; i++)
    {
        byteBuffer = pBytes[i];
        pBytes[i] = pBytes[stringLen-i-1];
        pBytes[stringLen-i-1] = byteBuffer;
    }
    return TRUE;
}
