/* file oneofexponent.c
 *
 * NIST PIV Demo CSP -- Cryptographic Service Provider
 *
 * Designed and Developed by Electrosoft (www.electrosoft-inc.com) based on requirements
 * from the National Institute of Standards and Technology (NIST).
 *
 * This file codes and documents all the necessary functions to import a session key.
 *
 * Adapted from an MSDN article
 */
#include <windows.h>
#include <wincrypt.h>
#include <cspdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <winscard.h>
#include "csp-debug.h"
#include "cspdefs.h"

#ifdef ISO24727
#include "iso24727api.h"
#else
#include "piv.h"
#endif

#include "csp.h"
#include "csputil.h"

BOOL ImportPlainSessionBlobModified(HCRYPTPROV hProv,
                            HCRYPTKEY hPrivateKey,
                            ALG_ID dwAlgId,
                            LPBYTE pbKeyMaterial ,
                            DWORD dwKeyMaterial ,
                            HCRYPTKEY *hSessionKey)
{
   BOOL fResult;
   BOOL fReturn = FALSE;
   BOOL fFound = FALSE;
   LPBYTE pbSessionBlob = NULL;
   DWORD dwSessionBlob, dwSize, n;
   DWORD dwPublicKeySize;
   DWORD dwProvSessionKeySize;
   ALG_ID dwPrivKeyAlg;
   LPBYTE pbPtr;
   DWORD dwFlags = CRYPT_FIRST;
   PROV_ENUMALGS_EX ProvEnum;
   HCRYPTKEY hTempKey = 0;

   unsigned char edata[3000];
   DWORD elen = dwKeyMaterial;


   __try
   {
		// Double check to see if this provider supports this algorithm
		// and key size
		do
		{
			dwSize = sizeof(ProvEnum);
			fResult = CryptGetProvParam(hProv, PP_ENUMALGS_EX, (LPBYTE)&ProvEnum,
									 &dwSize, dwFlags);
			if (!fResult)
			{
				DEBUG(3, "CryptGetProvParam failed!!\n");
				break;
			}
			dwFlags = 0;

			if (ProvEnum.aiAlgid == dwAlgId) fFound = TRUE;

		} while (!fFound);

		if (!fFound) __leave;

		// We have to get the key size(including padding)
		// from an HCRYPTKEY handle.  PP_ENUMALGS_EX contains
		// the key size without the padding so we can't use it.

		fResult = CryptGenKey(hProv, dwAlgId, 0, &hTempKey);
		if (!fResult) __leave;

		dwSize = sizeof(DWORD);
		fResult = CryptGetKeyParam(hTempKey, KP_KEYLEN, (LPBYTE)&dwProvSessionKeySize,
                                 &dwSize, 0);
		if (!fResult) __leave;
		CryptDestroyKey(hTempKey);
		hTempKey = 0;

		// Our key is too big, leave
		if ((dwKeyMaterial * 8) > dwProvSessionKeySize) __leave;

		// Get private key's algorithm
		dwSize = sizeof(ALG_ID);
		fResult = CryptGetKeyParam(hPrivateKey, KP_ALGID, (LPBYTE)&dwPrivKeyAlg, &dwSize, 0);
		if (!fResult) __leave;


		memcpy(edata, pbKeyMaterial, dwKeyMaterial);

		// we will encrypt with original key handle
		if (RCRYPT_FAILED(CryptEncrypt(hPrivateKey,
					0,
					(BOOL) TRUE,
					0,
					(BYTE *) edata,
					(DWORD *) &elen,
					3000)))
		{
			DEBUG(3, "CryptEncrypt Failed\n");
			__leave;
		}

		// Get private key's length in bits
		dwSize = sizeof(DWORD);
		fResult = CryptGetKeyParam(hPrivateKey, KP_KEYLEN, (LPBYTE)&dwPublicKeySize, &dwSize, 0);
		if (!fResult) __leave;

		// calculate Simple blob's length
		dwSessionBlob = (dwPublicKeySize/8) + sizeof(ALG_ID) + sizeof(BLOBHEADER);

		// allocate simple blob buffer
		pbSessionBlob = (LPBYTE)LocalAlloc(LPTR, dwSessionBlob);
		if (!pbSessionBlob) __leave;

		pbPtr = pbSessionBlob;

		// SIMPLEBLOB Format is documented in SDK
		// Copy header to buffer
		((BLOBHEADER *)pbPtr)->bType = SIMPLEBLOB;
		((BLOBHEADER *)pbPtr)->bVersion = 2;
		((BLOBHEADER *)pbPtr)->reserved = 0;
		((BLOBHEADER *)pbPtr)->aiKeyAlg = dwAlgId;
		pbPtr += sizeof(BLOBHEADER);

		// Copy private key algorithm to buffer
		*((DWORD *)pbPtr) = dwPrivKeyAlg;

		pbPtr += sizeof(ALG_ID);

		// Place the key material in reverse order
		for (n = 0; n < elen; n++)
		{
			pbPtr[n] = edata[n]; //edata[elen-n-1];
		}
		fResult = CryptImportKey(hProv, pbSessionBlob , dwSessionBlob,
						   hPrivateKey, CRYPT_EXPORTABLE, hSessionKey);
		if (!fResult)
		{
			DEBUG(3, " CryptImportKey failed with \n");
			__leave;
		}

		fReturn = TRUE;
   }
   __finally
   {
      if (hTempKey) CryptDestroyKey(hTempKey);
      if (pbSessionBlob) LocalFree(pbSessionBlob);
   }

   return fReturn;
}

BOOL CreateRealKey(LPTSTR szProvider,
                                 DWORD dwProvType,
                                 LPTSTR szContainer,
                                 DWORD dwKeySpec,
                                 HCRYPTPROV *hProv,
                                 HCRYPTKEY *hPrivateKey)
{
	unsigned char debug[100];

   BOOL fReturn = FALSE;
   BOOL fResult;
   LPBYTE keyblob = NULL;

   DEBUG(3, "Creating real key\n");

   __try
   {
      *hProv = 0;
      *hPrivateKey = 0;

      if ((dwKeySpec != AT_KEYEXCHANGE) && (dwKeySpec != AT_SIGNATURE))  __leave;

      // Try to create new container
      fResult = CryptAcquireContext(hProv, szContainer, szProvider,
                                    dwProvType, CRYPT_NEWKEYSET);
      if (!fResult)
      {
         // If the container exists, open it
         if (GetLastError() == NTE_EXISTS)
         {
            fResult = CryptAcquireContext(hProv, szContainer, szProvider, dwProvType, 0);
            if (!fResult)
            {
               // No good, leave
               __leave;
            }
         }
         else
         {
            // No good, leave
            __leave;
         }
      }

      // Generate the private key
      fResult = CryptGenKey(*hProv, dwKeySpec, CRYPT_EXPORTABLE, hPrivateKey);
      if (!fResult) __leave;

      fReturn = TRUE;
	  DEBUG(3, "Created normal key, no massaging whatsoever\n!!");
   }
   __finally
   {
	   DEBUG(3, "Executing finally block real key\n");
      if (keyblob) LocalFree(keyblob);

      if (!fReturn)
      {
         if (*hPrivateKey) CryptDestroyKey(*hPrivateKey);
         if (*hProv) CryptReleaseContext(*hProv, 0);
      }
   }
   sprintf(debug, "Returning from create real key %d\n", fReturn);
   DEBUG(3, debug);
   return fReturn;
}